import os
import time
from importlib import reload

import pytest

import uyeia


@pytest.fixture(autouse=True)
def cleanup_tests_folder():
    def delete_file_config(path):
        if os.path.exists(path):
            # Retry logic for Windows file locking issues
            max_attempts = 20
            for attempt in range(max_attempts):
                try:
                    os.remove(path)
                    break
                except (PermissionError, OSError):
                    if attempt < max_attempts - 1:
                        time.sleep(0.1)  # Wait 100ms before retry
                        continue

    delete_file_config("./tests/uyeia.errors.json")
    delete_file_config("./tests/errors_cache.db")
    delete_file_config("./tests/samples/corrupted_cache.db")
    yield
    delete_file_config("./tests/uyeia.errors.json")
    delete_file_config("./tests/errors_cache.db")
    delete_file_config("./tests/samples/corrupted_cache.db")


@pytest.fixture(autouse=True)
def reload_package():
    global uyeia
    reload(uyeia)
    yield


@pytest.fixture
def sample_config():
    return uyeia.Config(
        error_config_location="./tests/samples/uyeia.errors.json",
        error_cache_location="./tests/errors_cache.db",
    )


@pytest.fixture(autouse=True)
def ensure_db_cleanup():
    """Ensure database connections are properly closed after each test."""
    yield

    # Force close any remaining database connections
    if hasattr(uyeia, "__root__"):
        try:
            if hasattr(uyeia.__root__, "close_cache"):
                uyeia.__root__.close_cache()
        except Exception:
            pass

    # Additional Windows-specific cleanup with retry logic
    import gc

    gc.collect()  # Force garbage collection

    db_paths = ["./tests/errors_cache.db", "./tests/samples/corrupted_cache.db"]
    for db_path in db_paths:
        # Clean up both the main file and associated SQLite files
        for suffix in ["", "-wal", "-shm"]:
            file_path = db_path + suffix
            if os.path.exists(file_path):
                max_attempts = 10
                for attempt in range(max_attempts):
                    try:
                        os.remove(file_path)
                        break
                    except (PermissionError, OSError):
                        if attempt < max_attempts - 1:
                            time.sleep(0.05)  # Wait 50ms before retry
                            continue



import os
import sqlite3
import time

from uyeia.exceptions import UYEIACacheError
from uyeia.type import Status


def delete_file_db(path):
    if os.path.exists(path):
        # Retry logic for Windows file locking issues
        max_attempts = 20
        for attempt in range(max_attempts):
            try:
                os.remove(path)
                break
            except (PermissionError, OSError):
                if attempt < max_attempts - 1:
                    time.sleep(0.2)  # Wait 100ms before retry
                    continue


class UyeiaCache:
    """
    A simple cache class that uses SQLite to store watcher statuses.
    This class provides methods to set, get, and clear cache entries.
    It is designed to be thread-safe and can be used across multiple threads.
    """

    def __init__(self, db_location: str):
        self.db_location = db_location
        self.__connection = None
        self.__setup()

    def __connect(self):
        """Connect to the SQLite database, recreate if corrupted or missing."""
        if not self.__connection or not isinstance(
            self.__connection, sqlite3.Connection
        ):
            # Enable WAL mode for better concurrent access on Windows
            connection = sqlite3.connect(self.db_location, timeout=30.0)
            if os.name == "nt":
                connection.execute("PRAGMA journal_mode=WAL")
                connection.execute("PRAGMA synchronous=NORMAL")
                connection.execute("PRAGMA cache_size=10000")
                connection.execute("PRAGMA temp_store=MEMORY")
            self.__connection = connection
        return self.__connection

    def __setup(self, recall: bool = False):
        """Initialise the cache database (tables)."""
        cursor = None
        connection = None
        try:
            connection = self.__connect()
            cursor = connection.cursor()
            # Create the cache table
            cursor.execute("""
                CREATE TABLE IF NOT EXISTS cache (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    watcher TEXT NOT NULL,
                    status TEXT,
                    message TEXT,
                    solution TEXT,
                    escalation INTEGER,
                    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    updated_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    UNIQUE(watcher) ON CONFLICT REPLACE
                )
            """)

            # Create the vault table
            cursor.execute("""
                CREATE TABLE IF NOT EXISTS vault (
                    key TEXT PRIMARY KEY,
                    value TEXT NOT NULL,
                    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    updated_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    UNIQUE(key) ON CONFLICT REPLACE
                )
            """)

            connection.commit()
        except sqlite3.DatabaseError:
            if cursor:
                cursor.close()
            if self.__connection:
                self.__connection.close()
            if not recall:
                # Corrupted database → delete and rebuild
                self.__connection = None
                delete_file_db(self.db_location)
                # Reconnect and retry once
                self.__setup(recall=True)
            else:
                raise UYEIACacheError("Corrupted database could not be recovered")

    def __query(self, query: str, params: tuple = ()):
        connection = self.__connect()
        cursor = connection.cursor()
        try:
            cursor.execute(query, params)
            command = query.strip().split()[0].upper()

            # If it's a SELECT or similar, fetch results
            if command in ("SELECT", "PRAGMA", "EXPLAIN", "WITH"):
                result = cursor.fetchall()
            else:
                connection.commit()
                # Return useful info for non-SELECTs
                result = {"rowcount": cursor.rowcount, "lastrowid": cursor.lastrowid}
            return result
        except sqlite3.Error as e:
            raise UYEIACacheError(f"Cache query failed: {e}")
        finally:
            cursor.close()

    def exists(self, field: str, value: str) -> bool:
        query = f"SELECT 1 FROM cache WHERE {field} = ? LIMIT 1"
        params = (value,)

        try:
            result = self.__query(query, params)
            return bool(result)
        except UYEIACacheError as e:
            raise UYEIACacheError(f"Failed to check existence in cache: {e}")

    def set(self, watcher: str, status: Status):
        """Set a value in the cache."""
        query = """
            INSERT INTO cache (watcher, status, message, solution, escalation)
            VALUES (?, ?, ?, ?, ?)
        """
        params = (
            watcher,
            status["status"],
            status["message"],
            status["solution"],
            status["escalation"],
        )

        try:
            self.__query(query, params)
        except UYEIACacheError as e:
            raise UYEIACacheError(f"Failed to set cache: {e}")

    def get_all(self) -> dict[str, dict[str, Status]]:
        data = {}
        query = "SELECT watcher, status, message, solution, escalation FROM cache"
        try:
            result = self.__query(query)
            for row in result:
                data.setdefault(row[1], {}).setdefault(row[0], {}).update(
                    {
                        "watcher": row[0],
                        "status": row[1],
                        "message": row[2],
                        "solution": row[3],
                        "escalation": row[4],
                    }
                )
            return data
        except UYEIACacheError as e:
            raise UYEIACacheError(f"Failed to get all watchers from cache: {e}")

    def remove(self, watcher: str, status: Status):
        """Remove a watcher from the cache."""
        query = "DELETE FROM cache WHERE watcher = ? AND status = ?"
        params = (watcher, status["status"])

        try:
            self.__query(query, params)
        except UYEIACacheError as e:
            raise UYEIACacheError(f"Failed to remove watcher from cache: {e}")

    def get(self, field: str, value: str) -> Status | None:
        query = f"SELECT status, message, solution, escalation FROM cache WHERE {field} = ? LIMIT 1"
        params = (value,)

        try:
            result = self.__query(query, params)
            if result:
                status, message, solution, escalation = result[0]
                return {
                    "status": status,
                    "message": message,
                    "solution": solution,
                    "escalation": escalation,
                }
            return None
        except UYEIACacheError as e:
            raise UYEIACacheError(f"Failed to get watcher from cache: {e}")

    def escalate(self, status: list[str]):
        query = "UPDATE cache SET escalation = COALESCE(escalation, 0) + 1 WHERE status NOT IN (?)"
        params = (",".join(status),)
        try:
            self.__query(query, params)
        except UYEIACacheError as e:
            raise UYEIACacheError(f"Failed to escalate cache: {e}")

    def update_after_escalate(self, max_escalation: int, high_status: str):
        query = "UPDATE cache SET status = ? WHERE escalation >= ?"
        params = (high_status, max_escalation)
        try:
            self.__query(query, params)
        except UYEIACacheError as e:
            raise UYEIACacheError(f"Failed to update cache after escalate: {e}")

    def clear(self):
        """Clear the cache."""
        query = "DELETE FROM cache"
        try:
            self.__query(query)
        except UYEIACacheError as e:
            raise UYEIACacheError(f"Failed to clear cache: {e}")

    def close(self):
        """Close the database connection."""
        if getattr(self, "_UyeiaCache__connection", False):
            self.__connection.close()
            self.__connection = None

    def set_vault(self, key: str, value: str):
        query = "INSERT INTO vault (key, value) VALUES (?, ?)"
        params = (key, value)
        try:
            self.__query(query, params)
        except UYEIACacheError as e:
            raise UYEIACacheError(f"Failed to set vault: {e}")

    def remove_vault(self, key: str):
        query = "DELETE FROM vault WHERE key = ?"
        params = (key,)
        try:
            self.__query(query, params)
        except UYEIACacheError as e:
            raise UYEIACacheError(f"Failed to remove vault: {e}")

    def retrieve_vault(self, key: str) -> str | None:
        query = "SELECT value FROM vault WHERE key = ? LIMIT 1"
        params = (key,)
        try:
            result = self.__query(query, params)
            if result:
                return result[0][0]
            return None
        except UYEIACacheError as e:
            raise UYEIACacheError(f"Failed to retrieve vault: {e}")

    def __enter__(self):
        """Context manager entry."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit - ensure connection is closed."""
        self.close()

    def __del__(self):
        """Close the database connection when the cache is deleted."""
        self.close()



import uyeia


def test_put_vault(sample_config):
    uyeia.set_global_config(sample_config)
    vault = uyeia.Vault()
    vault.set("test_key", "test_value")
    value = vault.get("test_key")
    assert value == "test_value"


def test_modify_vault(sample_config):
    uyeia.set_global_config(sample_config)
    vault = uyeia.Vault()
    vault.set("test_key", "init_value")
    vault.set("test_key", "new_test_value")
    value = vault.get("test_key")
    assert value == "new_test_value"


def test_remove_vault(sample_config):
    uyeia.set_global_config(sample_config)
    vault = uyeia.Vault()
    vault.set("test_key", "test_value")
    vault.remove("test_key")
    value = vault.get("test_key")
    assert value is None



import pytest

import uyeia


def test_global_config_unknow_logging_level_int():
    config = uyeia.Config(
        status={
            "HEALTHY": -1,
        }
    )

    with pytest.raises(uyeia.UYEIAConfigError):
        uyeia.set_global_config(config)


def test_global_config_unknow_logging_level_str():
    config = uyeia.Config(
        status={
            "HEALTHY": "test",
        }
    )

    with pytest.raises(uyeia.UYEIAConfigError):
        uyeia.set_global_config(config)



from __future__ import annotations

import atexit
import io
import json
import logging
import os
import re
import threading
from datetime import datetime, timezone
from typing import Literal

from uyeia.cache import UyeiaCache
from uyeia.exceptions import UYEIAConfigError
from uyeia.type import CommonStatus, Config, Status

__all__ = ["Watcher", "set_global_config", "get_errors", "Vault"]

_lock = threading.RLock()
_uyeia_config: Config = Config()
__uyeia_init__ = False
__default_logger__ = logging.getLogger("__UYEIA__")
__error_mapper__: dict[str, CommonStatus] = {}


def set_global_config(config: Config):
    global _uyeia_config, __root__

    if error := config.validate():
        raise UYEIAConfigError(f"Invalid UYEIA config: {error}")

    _uyeia_config = config
    __root__._init_cache()
    _init_uyeia_env()


def _load_error_definitions():
    path = _uyeia_config.error_config_location
    if os.path.isfile(path) and os.access(path, os.R_OK):
        try:
            with io.open(path, "r") as db_file:
                data = json.load(db_file)
                return data
        except json.decoder.JSONDecodeError as e:
            raise UYEIAConfigError("Invalid errors config:", e)

    with io.open(path, "w") as db_file:
        json.dump({}, db_file)
    return {}


def _init_uyeia_env():
    global __uyeia_init__, __error_mapper__, __root__

    with _lock:
        __error_mapper__ = _load_error_definitions()
        __uyeia_init__ = True


class Watcher:
    def __init__(self, name: str | None = None, logger: logging.Logger | None = None):
        self._cache: Status | None = None

        if logger:
            self.logger = logger
            self.name = self.logger.name
        elif name:
            self.logger = logging.getLogger(name)
            self.name = name
        else:
            raise ValueError("Name or logger is required for Watcher instance")

        global __root__
        if not __uyeia_init__:
            _init_uyeia_env()

        self.manager = __root__.register(self)

    def __log(self, status: str, message: str, config: Config):
        level = config.status.get(status)
        if isinstance(level, str):
            level = logging.getLevelName(level)

        if level:
            self.logger.log(level, message)
        else:
            raise ValueError(f"Invalid status: {status}. Not in UYEIA config!")

    def __is_empty_or_high(self, status: CommonStatus) -> bool:
        levels = list(_uyeia_config.status.keys())
        return not self._cache or levels.index(status["status"]) > levels.index(
            self._cache["status"]
        )

    def get_actual_status(self):
        return self._cache

    def __replace_vars(self, message, args):
        var = re.search(r"{{(.*?)}}", message)
        if not var:
            return message
        var_name = var.groups()[0]
        value = args.get(var_name, "")
        return self.__replace_vars(
            message[: var.start()] + str(value) + message[var.end() :], args
        )

    def register(self, error_code: str, custom_message=None, vars: dict[str, str] = {}):
        error = __error_mapper__.get(error_code)
        if not error:
            raise ValueError(
                f"Invalid error code: {error_code}. Not in UYEIA errors config!"
            )

        with _lock:
            if self.__is_empty_or_high(error):
                if self._cache and self._cache["status"] != error["status"].upper():
                    self.manager.delete_entry_cache(self.name, self._cache)

                message = custom_message or error["message"]
                if vars:
                    message = self.__replace_vars(message, vars)

                if not _uyeia_config.disable_logging:
                    self.__log(error["status"], message, _uyeia_config)
                self._cache = {
                    "status": error["status"].upper(),
                    "message": self.__add_timestamp(message),
                    "solution": error.get("solution", _uyeia_config.default_solution),
                    "escalation": 0,
                }
                self.manager.write_entry_cache(self.name, self._cache)

    def release(self):
        if not self._cache:
            return

        with _lock:
            self.manager.delete_entry_cache(self.name, self._cache)
            self._cache = None

    def __del__(self):
        if getattr(self, "manager", None):
            self.manager.unregister(self.name)

    def __add_timestamp(self, error_log: str) -> str:
        return (
            f"{datetime.now(timezone.utc).strftime('%Y-%m-%d %H:%M:%S')} - {error_log}"
        )


class Manager:
    def __init__(self) -> None:
        self.watcherDict: dict[str, Watcher] = {}

    def _init_cache(self):
        global _uyeia_config
        self.__cache = UyeiaCache(_uyeia_config.error_cache_location)

    def getWatcher(self, name: str | None = None, logger: logging.Logger | None = None):
        if name and logger:
            raise ValueError(
                "Name or logger is required for Watcher instance! Not both."
            )

        rv = None
        name = name or getattr(logger, "name", None)
        if not name:
            raise ValueError("Name or logger is required for Watcher instance")

        with _lock:
            if name in self.watcherDict:
                rv = self.watcherDict[name]
            else:
                rv = Watcher(name, logger)
        return rv

    def __find_data_watcher(self, watcher_name: str):
        with _lock:
            return self.__cache.get("watcher", watcher_name)

    def delete_entry_cache(self, name: str, old_status: Status):
        with _lock:
            self.__cache.remove(name, old_status)

    def write_entry_cache(self, name: str, status: Status):
        with _lock:
            self.__cache.set(name, status)

    def get_cache(self):
        return self.__cache

    def set_cache(self, new_cache):
        with _lock:
            self.__cache = new_cache
            for watcher in self.watcherDict.values():
                watcher._cache = self.__find_data_watcher(watcher.name)

    def clear_cache(self):
        with _lock:
            for watcher in self.watcherDict.values():
                watcher.release()

    def close_cache(self):
        with _lock:
            if getattr(self, "_Manager__cache", None):
                self.__cache.close()

    def register(self, watcher: Watcher):
        with _lock:
            self.watcherDict[watcher.name] = watcher
            watcher._cache = self.__find_data_watcher(watcher.name)
            return self

    def unregister(self, name: str):
        with _lock:
            self.watcherDict.pop(name, None)


__root__ = Manager()


class Vault:
    def __init__(self):
        self.__vault = __root__.get_cache()

    def set(self, key: str, value: str):
        if not self.__vault:
            return
        self.__vault.set_vault(key, value)

    def get(self, key: str) -> str | None:
        if not self.__vault:
            return None
        return self.__vault.retrieve_vault(key)

    def remove(self, key: str):
        if not self.__vault:
            return
        self.__vault.remove_vault(key)


def get_errors(mode: Literal["all", "hot", "cold"] = "all"):
    cache = __root__.get_cache()
    if not cache:
        return None

    if mode == "all":
        return cache.get_all()

    if mode not in {"hot", "cold"}:
        raise ValueError(f"Invalid mode: {mode}. Must be 'all', 'hot' or 'cold'!")

    for status in (
        _uyeia_config.status.keys()
        if mode == "cold"
        else reversed(_uyeia_config.status.keys())
    ):
        if cache.exists("status", status):
            return cache.get("status", status)

    return None


def escalate():
    if _uyeia_config.disable_escalation:
        __default_logger__.warning("Escalate function is disabled in config.")
        return

    cache = __root__.get_cache()
    with _lock:
        high_status = _uyeia_config.escalation_status or next(
            reversed(_uyeia_config.status)
        )
        first_status = next(iter(_uyeia_config.status))

        cache.escalate([high_status, first_status])
        cache.update_after_escalate(_uyeia_config.max_escalation, high_status)


atexit.register(__root__.close_cache)



import os
import shutil

import uyeia


def test_read_cache_error():
    config = uyeia.Config(
        error_cache_location="./tests/samples/errors_cache.db",
        error_config_location="./tests/samples/uyeia.errors.json",
    )
    uyeia.set_global_config(config)
    errors = uyeia.get_errors()
    assert errors and len(errors) == 2


def test_open_corrupted_cache():
    # TODO:  Windows permission issue
    if os.name == "nt":
        return

    import tempfile

    # Use a temporary file to avoid Windows file locking issues
    with tempfile.NamedTemporaryFile(suffix=".db", delete=False) as temp_db:
        temp_db_path = temp_db.name

    # Copy corrupted cache to temporary location
    shutil.copy("./tests/samples/.corrupted_cache.db", temp_db_path)

    config = uyeia.Config(
        error_cache_location=temp_db_path,
        error_config_location="./tests/samples/uyeia.errors.json",
    )
    uyeia.set_global_config(config)
    watcher = uyeia.Watcher("test_watcher")
    watcher.register("T404")
    error = uyeia.get_errors("hot")
    assert error and error.get("status") == "WARNING"
